#ifndef CAMERA_H
#define CAMERA_H

#include "IOmap.h"
#include "WPILib.h"
#include "ThunderVision.h"

// A whole slew of #defines:

		/* Targets */
// actual height of target, in inches
#define H0 18
// actual width of target, in inches
#define W0 24
// canned for a slight calculation time improvement
#define TWO_FH0 (2/F/H0)
#define FW0 F*W0
// target center heights (hoop height+offset+half target height-camera height)
#define TOP_HEIGHT 98+2+H0/2-CAM_HEIGHT
#define CENTER_HEIGHT 61+2+H0/2-CAM_HEIGHT
#define BOTTOM_HEIGHT 28+2+H0/2-CAM_HEIGHT
// Y positions, for RPS, of target centers (inches)
#define CENTRAL_YPOS (27*12)/2
#define LEFT_YPOS CENTRAL_YPOS+27.325
#define RIGHT_YPOS CENTRAL_YPOS-27.325
// horizontal distance from center of target to center of basket (inches)
#define T_B_DIST 6+18/2

		/* Targets in image */
// minimum pixel height and width of a prospective target; anything less in
// either dimension and it's probably noise
#define MIN_WIDTH 30
#define MIN_HEIGHT 20
// maximum distance target center can be from image center (in pixels)
// and still be counted as "centered"; TODO tune!
#define MAX_DELTA_X 5
// max number of degrees that a target can be tilted from the horizontal (either
// direction) before we say it's tilted. TODO tune!
#define TILT_TOL 0.5
// if a target's center is EXTREME_RATIO times its width to left/right of another's,
// we can consider them to be a center_left and center_right pair. The same argument
// with height and above/below will identify top and bottom pairs. TODO tune!!
#define EXTREME_RATIO 1.8
// like EXTREME_RATIO, but this distinguishes top/center_left pairs from
// center_left/center_right pairs (i.e., is the target really above the other or
// just perspective-distorted); same for horizontal differences
#define MEAN_RATIO EXTREME_RATIO/2.0

		/* Camera */
// distance from camera to center of shooter TODO check this value!
#define CAM_SHOOTER_DIST 3.1
// camera height above floor (in.) TODO check it!
#define CAM_HEIGHT 38
// difference between camera servo value and actual camera angle from horizontal
// TODO tune!
#if ROBOT_2
#define CAM_ANGLE_CORR 60.0
#else
#define CAM_ANGLE_CORR 65.0
#endif
// calibration points: distance target to camera (horizontal, no
//  height difference, target perpendicular to camera line of sight);
// then target height, in pixels, under given conditions TODO improve?
#define CALIB_DISTANCE 60.0
#define CALIB_TGT_HT_PX 235.0
// effective focal length of camera [(d*h_image)/h_0]
#define F (CALIB_DISTANCE * CALIB_TGT_HT_PX / H0)
// camera image size and midpoints
#define IMAGE_WIDTH 640
#define IMAGE_HEIGHT 480
#define IMAGE_CENTER_H IMAGE_WIDTH/2
#define IMAGE_CENTER_V IMAGE_HEIGHT/2

class Camera {
public:
	// identity of our main target (the one used for position/shot calculations)
	// center-left/right are as seen by the robot from the field
	/* "center" is for driver use only (currently, they can only specify top, center,
	 *  or bottom target), and I hope I'll be able to drop it in future TODO
	 * Camera pretty much just ignores the TgtID "center" */
	typedef enum TargetIdentity { none, top, center_left, center_right, bottom, center };
	struct Target {
		int c_x; // center point x (pixels)
		int c_y; // center point y (pixels)
		float h; // height in pixels (+fractional pixels)
		float w; // width in pixels
		int tilt; // -1: left side lower; 0: flat; +1: right side lower
	};
	
	Camera(); //constructor
	~Camera();
	
	/* QUERY: split this into GetNew and Analyze? */
	// get an image (or 2?), analyze it (them?), and store the relevant data
	void GetNewImage();
	
	// return the height and width of the target (in pixels)
	float HeightOfTarget();
	float WidthOfTarget();
	float TargetCenterHeight(); // height of target center above ground
	float TargetCenterYPos(); // for RPS: position of target center in y-plane
	// distance of center of target from center of image (horizontally, in pixels)
	// NOTE: left is positive, right is negative
	int TargetHorizDelta();
	int TargetTilt(); // tilt of main target (-1:right side higher, 0:flat, +1:left up)
	/** Note on target tilt: When robot is to right of target, its y position will be lower than
	 * the target's, i.e. delta_y<0. In the camera image, left side of target will be higher when
	 * the robot is to right of target. Therefore, target tilted left side up denotes a negative
	 * delta_y from target to robot; so for convenience left side higher should be tilt +1.
	 * TL;DR: robot to right->tilt=+1; robot to left->tilt=-1. */
	
	// TgtFnd not much used; keep it around just in case
	bool TargetFound(); // was a target found?
	bool TargetIdentified(); // do we know which one it is?
	bool TargetCentered(); // is it close enough to center?

	// tilts camera to {angle} degrees from the horizontal
	void SetTilt(float angle);
	// returns current angle of camera (in degrees from horiz.)
	float GetTilt();
	// returns whether camera has reached desired angle yet
	/* ARE WE DUPLICATING GetTilt WITH THIS...??? */
	bool TiltDone();
private:
	// given all targets found by ImageAnalyzer, finds which one is closest to the image
	// center, and tries to determine which of the real world targets it is
	// parameter: number of targets found by IA->ProcessImage()
	/* TODO how many targets are we willing to allow? 4 only? and how do we know
	 *      which to exclude if we have more than the desired 4?? */
	void IDTargets(int count);
	// compares x and y position of targets[t1] and targets[t2] and returns the calculated
	// identity of the first using EXTREME_RATIO test, or "none" if the test is inconclusive
	TargetIdentity CompareTargets(int t1, int t2);
	
	// do we need to tilt? apparently...
	Servo *tilt;
	// no camera object?!
	ThunderVision *tv; // let's watch some TV! (actually it just analyzes images for us)
	
	// image gotten from the camera 
	Image *cam_image;
	// the array in which the filtered image is immediately stored
	RGBValue *image;

	// target data
	bool target_found; // have we found the "real" target?
	TargetIdentity best_t_id; // what is this "real" target you speak of?
	Target targets[5]; // the targets (at most 4, plus a buffer) we've found in the image
	                   // most central target is always at index 0
//	static const Target t_null; // null target. it's not a target at all.
	
	float camAngle; // angle camera is currently at
};

#endif
