#include "Shooter.h"
#include "Storage.h"
#include "IOmap.h"
#include <math.h>

// PID LOOPS *******************************

#if ROBOT_2
#define WHEEL_P_TOP			0.5f
#define WHEEL_I_TOP			0.008f
#define WHEEL_D_TOP			0.0f
#define WHEEL_P_BOT			0.5f
#define WHEEL_I_BOT			0.008f
#define WHEEL_D_BOT			0.0f
#define WHEEL_PID_PERIOD		0.005 // seconds
#define WHEEL_SPEED_SAMPLES 25//3
#define WHEEL_TOLERANCE_TOP  4   // %? of total. PIDController code used
#define WHEEL_TOLERANCE_BOT  5   // %? of total. PIDController code used
#else
#if FLR
#define WHEEL_P_TOP			7.5
#define WHEEL_I_TOP			.8
#define WHEEL_D_TOP			.2
#define WHEEL_P_BOT			7.5
#define WHEEL_I_BOT			.8
#define WHEEL_D_BOT			.2
#define WHEEL_PID_PERIOD		0.05 // seconds
#define WHEEL_SPEED_SAMPLES 1
#define WHEEL_TOLERANCE  1   // %? of total. PIDController code used
#else
#define WHEEL_P_TOP			0.8f
#define WHEEL_I_TOP			0.012
//#define WHEEL_I_TOP			0.0
//#define WHEEL_D_TOP			0.05
#define WHEEL_D_TOP			0.0
#define WHEEL_P_BOT 	0.5f //0.38f
#define WHEEL_I_BOT		0.008
#define WHEEL_D_BOT		0.0
#define WHEEL_PID_PERIOD		0.005 // seconds
#define WHEEL_SPEED_SAMPLES 25
#define WHEEL_TOLERANCE_TOP  4   // %? of total. PIDController code used
#define WHEEL_TOLERANCE_BOT  5   // %? of total. PIDController code used
#endif// FLR
#endif//ROBOT_2
#define TURRET_P		5
#define TURRET_I		0.1
#define TURRET_D		0
#define TURRET_PID_PERIOD		0.05 // seconds
#define TURRET_TOLERANCE (0.03/4)  // % of total expressed as decimal. Own tolerance
                               // code used

// LIMITS ********************************

#define SHOOTER_MAX_RPS 59// NEEDS TO BE HIGHER
#define MINIMUM_SHOOT_SPEED .10 // of total; to prevent clogups due to shooting with no speed
#define MINIMUM_SHOOT_POWER .15

#define USE_SCALE_BACKSPIN 1
#if USE_SCALE_BACKSPIN
#define MAX_SPEED_NO_BACKSPIN 0.80	// all speeds above this will have no backspin
#define MIN_SPEED_FULL_BACKSPIN 0.40 // all speeds below this will have FULL_BACKSPIN 
#define FULL_BACKSPIN 0.5
#else
#define MIN_USE_SPEED_FACTOR 0.40 // if speed less than this, use 'hard coded 
#define TOP_SPEED_FACTOR 0.5// For use in giving more pronounced backspin to the shooter wheels
#define BOTTOM_SPEED_FACTOR 1
#endif

// TURRET ********************************

//max encoder counts for our range of motion - what our encoder should read if
//  we hit the max limit switch
//(1261 * 4)
#if ROBOT_2
#define TURRET_MAX_COUNTS 5108
#else
#define TURRET_MAX_COUNTS (3628)
#endif
//difference between driver "0" (front of bot, intake side) vs robot "0" (midpoint between stops)
//  we move the driver setpoint into robot degrees so we can easily scale the PID input
//  from -1 to +1 for errors between min and max encoder counts
//  This should be the "driver degrees" equivalent of halfway through the range of motion
#define TURRET_OFFSET_DEG (((TURRET_MAX_DEG - TURRET_MIN_DEG) / 2.0) + TURRET_MIN_DEG)


// ENCODERS *******************************

// For shooter wheels
#define TICKS_PER_REV   250


// TIME PAUSES ********************************

#define SHOOTING_SLOWDOWN_DELAY 1// second(s)
#define MIN_SHOOT_REST 		2*WHEEL_PID_PERIOD

// CODE!! 

static double turretNormalizeTicks(int ticks);
static double turretCountToAngle(int count);
static double turretAngleToNormalizedTicks(double angle);


//----------------- MONO ENCODER -------------------------------------

MonoEncoder::MonoEncoder(int channel) {
	count = new Counter(channel);
	timer = new Timer();
	
	count->Start();
	timer->Start();
	totalCount = 0;
	lastSpeed = 0;
}

MonoEncoder::~MonoEncoder() {
	delete count;
	delete timer;
}

int MonoEncoder::getTotalCount() {
    return totalCount + count->Get();
}

double MonoEncoder::getLastSpeed() {
	return lastSpeed;
}

double MonoEncoder::getRevPerSec() {
	return lastSpeed * (double)SHOOTER_MAX_RPS;
}

double MonoEncoder::PIDGet() {
	double sec;
	double ret;
	int c;
	//50ms update rate. 38% = 200 counts, 60% = 350 counts
//	float P,I,D;
    timer->Stop();// is this needed?
    sec = timer->Get();
    count->Stop();
    c = count->Get();
    
    ret = c;

    // ticks * rev/ticks / time = revs/sec
    ret /= TICKS_PER_REV;   // ticks / (ticks per rev) = revs
    //printf("ret tpr; %f\n", ret);
    ret /= sec;             // revs / seconds = revs / second
    totalCount += c;
    
    ret /= SHOOTER_MAX_RPS; // Normalize for PID controller
    
    // PID return
    // print out current P I and D values for our arrows
    
    
    lastSpeed = ret;   
    
    timer->Reset();
    count->Reset();
    timer->Start();
    count->Start();
    return ret;
}

// ---------------------  LIMITED OUTPUT -----------------------

LimitedOutput::LimitedOutput(PIDOutput *out, AnalogChannel *fwdLimit, AnalogChannel *revLimit)
{
	this->output = out;
	this->fwdLimit = fwdLimit;
	this->revLimit = revLimit;
}
void LimitedOutput::PIDWrite(float f) {
    f = -f;
	if (f >= 0.0f && fwdLimit->GetVoltage() < 2) {
		f = 0;
	} else if (f <= 0.0f && revLimit->GetVoltage() < 2) {
		f = 0;
	}
	output->PIDWrite(f);
}


class TurretEncoder : public Encoder {
public:
    TurretEncoder(int a, int b) : Encoder(a, b) {
    }
    
    virtual double PIDGet() {
        return turretNormalizeTicks(this->GetRaw()); 
    }
};


//------------------------- SHOOTER CLASS --------------------

Shooter::Shooter(Storage * s)
{
	store = s;

	// Makes three PID encoders for the rotary units
	turret_jaguar = new ThunderCANJaguar(CANID_TURRET);
	top_jaguar = new ThunderCANJaguar(CANID_SHOOTER_UPPER);
	bottom_jaguar = new ThunderCANJaguar(CANID_SHOOTER_LOWER);
	
	turret_encoder = new TurretEncoder(DIG_ENCODER_TURRET_B, DIG_ENCODER_TURRET_A);
	turret_encoder->SetReverseDirection(true);
	top_encoder = new MonoEncoder(DIG_ENCODER_SHOOTER_UPPER);
	bottom_encoder = new MonoEncoder(DIG_ENCODER_SHOOTER_LOWER);

	turret_encoder->Start();
	//turret_encoder = new Encoder(DIG_ENCODER_TURRET_A, DIG_ENCODER_TURRET_B);
	//top_encoder = new Encoder(DIG_ENCODER_SHOOTER_UPPER_A, DIG_ENCODER_SHOOTER_UPPER_B);
	//bottom_encoder = new Encoder(DIG_ENCODER_SHOOTER_LOWER_A, DIG_ENCODER_SHOOTER_LOWER_B);
	
	limitCW = new AnalogChannel(ANA_TURRET_SW_MAX_CW);
	limitCCW = new AnalogChannel(ANA_TURRET_SW_MAX_CCW);
	
	turretOutput = new LimitedOutput(turret_jaguar, limitCW, limitCCW);
	
	PIDOutput *top_output = top_jaguar;
    PIDOutput *bottom_output = bottom_jaguar;

#if SHOOTER_LOGGING
    top_logger = new UDPLogger();
    top_logger->init(TOP_LOGGER_PORT);
    top_logger->setEnabled(false);
    bottom_logger = new UDPLogger();
    bottom_logger->init(BOTTOM_LOGGER_PORT);
    bottom_logger->setEnabled(false);
    top_logger_output = new LoggingPIDOutput(top_jaguar, top_logger, top_encoder);
    bottom_logger_output = new LoggingPIDOutput(bottom_jaguar, bottom_logger, bottom_encoder);
    top_output = top_logger_output;
    bottom_output = bottom_logger_output;
#endif
	
	turret_PID = new PIDController(TURRET_P, TURRET_I, TURRET_D, turret_encoder, turretOutput, TURRET_PID_PERIOD);
    turret_PID->SetInputRange(-1.0, 1.0);
    turret_PID->SetOutputRange(-1.0, 1.0);
    turret_PID->SetTolerance(TURRET_TOLERANCE);
    
	top_PID = new PIDController(WHEEL_P_TOP,WHEEL_I_TOP,WHEEL_D_TOP, top_encoder, top_output, WHEEL_PID_PERIOD);
	top_PID->SetInputRange(0, 1.0);
    top_PID->SetTolerance(WHEEL_TOLERANCE_TOP);
	top_PID->SetOutputRange(0, 1.0);
	top_PID->Disable();
	
	bottom_PID = new PIDController(WHEEL_P_BOT,WHEEL_I_BOT,WHEEL_D_BOT, bottom_encoder, bottom_output, WHEEL_PID_PERIOD);
    bottom_PID->SetInputRange(0, 1.0);
    bottom_PID->SetOutputRange(0, 1.0);
    bottom_PID->SetTolerance(WHEEL_TOLERANCE_BOT);
	bottom_PID->Disable();
	
#if SHOOTER_LOGGING
    top_logger_output->setController(top_PID);
	bottom_logger_output->setController(bottom_PID);
#endif

	shot_timer = new Timer();
   
    speedup_timer = new Timer();

	// turret variables
	turret_current_angle = 0;
	turret_to_position = true;
	turret_rotvel = 0;
	turret_target_angle = 0;
	turret_move_manual = true;
	broken_turret = false;
	want_zero = false;
	// shooting variables
	shooting = false;// is the robot shooting ?
	want_shooting = false; // will the robot try to shoot ?
	// wheel variables
	top_speed = 0; // the speed the top wheel should attain
	bottom_speed = 0;// speed for the bottom wheel to attain
	top_at_speed = 0;
	bottom_at_speed = 0;
	
	shooter_enabled = false;// if the shooter wheels are spinning
							// enabled when the robot needs to shoot
							// disabled when not
	speed_variable_mode = true;
	wheels_broken = false;
	
	useBackspin = false;
}

Shooter::~Shooter() {
    turret_PID->Disable();
    delete turret_PID;
    
    top_PID->Disable();
    bottom_PID->Disable();
        
    delete top_PID;
    delete bottom_PID;

#if SHOOTER_LOGGING
    delete top_logger_output;
    delete bottom_logger_output;
    delete top_logger;
    delete bottom_logger;
#endif
    
	delete speedup_timer;

	delete turret_encoder;
	delete bottom_encoder;
	delete top_encoder;
	delete turretOutput;
	delete limitCW;
	delete limitCCW;

    delete turret_jaguar;
	delete top_jaguar;
	delete bottom_jaguar;

	delete shot_timer;

}

// --------------------- SHOT --------------------------------

void Shooter::shoot() {
	if (!shooting) // Don't start shooting again if we already were
	    want_shooting = true; // we're shooting now!
}

bool Shooter::shootFinished() {
	return (!shooting && !want_shooting);
}

void Shooter::shootStop() {
	want_shooting = false; // cancel that!
	shooting = false;
	shot_timer->Reset();
	store->cancelDeliverBall();
}

// --------------------------------- PROCESS -----------------------------------------------

void Shooter::processShooter() {
	// enable wheels if shooting
	
	if ((want_shooting && !shooter_enabled) || wheels_broken) shooterEnable();
	if (!want_shooting && !shooting && shooter_enabled && !wheels_broken) shooterDisable();	

	
	
#if SHOOTER_DEBUG
	static int throttle = 0;
	if ((throttle++ % 20) == 0) {
	printf("TCurrent PID output: %f setpoint: %f last measure: %f error: %f\n", (double)top_PID->Get(), (double)top_PID->GetSetpoint(), (double)top_encoder->getLastSpeed(), (double)top_PID->GetError());
	printf("BCurrent PID output: %f setpoint: %f last measure: %f error: %f\n", (double)bottom_PID->Get(), (double)bottom_PID->GetSetpoint(), (double)bottom_encoder->getLastSpeed(), (double)bottom_PID->GetError());
//	printf("B Power/Speed: %f\n", bottom_speed);
	}

#endif
	
	if (want_shooting && !(speedup_timer->Get() > 0)) {
		// The speedup_timer ensures that the PID loop has a chance to update
		// before the check for speed is called; this makes the different update rates
		// of the main code and the PID loop not cause problems. Without this,
		// occasionally the ball shoots instantaneously at the wrong speed
		speedup_timer->Start();
	}
		
	// if the call for shooting has been raised and one can shoot, shoot 
	if (want_shooting && atSpeed() && turretFinished() && speedup_timer->Get() > MIN_SHOOT_REST) {
#if SHOOTER_DEBUG
		printf("Time needed to start shooting; %f\n", speedup_timer->Get());
		printf("Sh0TCurrent PID output: %f setpoint: %f last measure: %f error: %f\n", (double)top_PID->Get(), (double)top_PID->GetSetpoint(), (double)top_encoder->getLastSpeed(), (double)top_PID->GetError());
		printf("Sh0BCurrent PID output: %f setpoint: %f last measure: %f error: %f\n", (double)bottom_PID->Get(), (double)bottom_PID->GetSetpoint(), (double)bottom_encoder->getLastSpeed(), (double)bottom_PID->GetError());

#endif
		speedup_timer->Stop();
		speedup_timer->Reset();
		
		shooting = true;
	    // start the delay to accommodate the time between balls and keeping
		// shooter wheels at desired speed throughout shot.
        shot_timer->Start();
		
		want_shooting = false;
		
		store->deliverBall();
#if SHOOTER_LOGGING
		double cur_time = GetTime();
    	printf("ball delivered %f\n", cur_time - at_speed_time);
    	at_speed_time = cur_time;
#endif
		
#if 0 && (TURRET_DEBUG || SHOOTER_DEBUG)
	} else if (want_shooting) {
		static int throttle2 = 0;
		if ((throttle2++ % 20) == 0)
	    printf("Waiting on turret? %d speed? %d\n", turretFinished(), atSpeed());
#endif
	}
	
	// declare shooting phase completed
	if (shooting && shot_timer->Get() > SHOOTING_SLOWDOWN_DELAY) {
		// ^ if enough time has passed after the ball exited
		shooting = false;
		shot_timer->Stop();
		shot_timer->Reset();
		// pid will be stopped next call if all other "don't spin anymore"
		// parameters are met

		// need to wait again for the wheels to get to speed
#if SHOOTER_LOGGING
		double cur_time = GetTime();
    	printf("at_speeds reset %f\n", cur_time - at_speed_time);
    	at_speed_time = cur_time;
#endif
		top_at_speed = 0;
    	bottom_at_speed = 0;
	}
	
	processTurret();
}

//----------------------------- SHOOTER --------------------------------------

void Shooter::setSpeed(float top, float bottom) {
	setNormalizedSpeed(top / SHOOTER_MAX_RPS, bottom / SHOOTER_MAX_RPS);
}

void Shooter::setNormalizedSpeed(float top, float bottom)
{
	// if the speed changed, then re-wait for it to be at that speed
	if (top != top_speed)
	{
		// make sure the change is more than 1 (slider jidders)
		if ((top > top_speed+1) || (top < top_speed-1))
			top_at_speed = 0;
	}
	if (bottom != bottom_speed)
	{
		if ((bottom > bottom_speed+1) || (bottom < bottom_speed-1))
			bottom_at_speed = 0;
	}
	
#if USE_SCALE_BACKSPIN
	if (useBackspin && (top < MAX_SPEED_NO_BACKSPIN))
	{
		if (top < MIN_SPEED_FULL_BACKSPIN) {
			top = top*FULL_BACKSPIN;
		} else {
			top *= ((top - MIN_SPEED_FULL_BACKSPIN)/(MAX_SPEED_NO_BACKSPIN - MIN_SPEED_FULL_BACKSPIN)*(1-FULL_BACKSPIN))+FULL_BACKSPIN;
		}
	}
#else
	if (top < MIN_USE_SPEED_FACTOR)// If the speed is less than this factor, induce backspin
	{
		top = top*TOP_SPEED_FACTOR;
		bottom = bottom*BOTTOM_SPEED_FACTOR;
	}
#endif
	
	// Limit the speed/power to a maximum of 100% for the wheels
	top_speed = top;
	if(top > 1.0f)
		top_speed = 1.0f;
	bottom_speed = bottom;
	if(bottom > 1.0f)
		bottom_speed = 1.0f;
	
	if (!shooter_enabled && speed_variable_mode) return;
	
    if(!wheels_broken)
    {
    	if (top_speed < MINIMUM_SHOOT_SPEED)
    		top_speed = MINIMUM_SHOOT_SPEED;
    	if (bottom_speed < MINIMUM_SHOOT_SPEED)
    		bottom_speed = MINIMUM_SHOOT_SPEED;
    	
		top_PID->SetSetpoint(top_speed);		    
	    bottom_PID->SetSetpoint(bottom_speed);
    } else {
    	//printf("manual jag in normalizespeed\n");

    	if (top_speed < MINIMUM_SHOOT_POWER)
    		top_speed = MINIMUM_SHOOT_POWER;
    	if (bottom_speed < MINIMUM_SHOOT_POWER)
    		bottom_speed = MINIMUM_SHOOT_POWER;
    	
    	bottom_jaguar->Set(bottom_speed);
    	top_jaguar->Set(top_speed);
    }
}

void Shooter::setWheelsBroken(bool broken)
{
	if (!wheels_broken && broken && (top_PID->IsEnabled() || bottom_PID->IsEnabled())) {
	    top_PID->Disable();
	    bottom_PID->Disable(); 
	    printf("Disabling shooter wheels\n");
        bottom_jaguar->Set(0);
        top_jaguar->Set(0);
	}
	wheels_broken = broken;
}

void Shooter::setWheelsVariable(bool variable) {
	speed_variable_mode = variable;// set
}

void Shooter::setUseBackspin(bool enable)
{
	useBackspin = enable;
}

bool Shooter::atSpeed()
{
    if (wheels_broken) {
        return true;
    }
    // need to be OnTarget for WHEEL_SPEED_SAMPLES.
    // once this has happened, then we don't care about
    // minor spikes.
	if (top_PID->OnTarget()) {
		top_at_speed++;
	} else if (top_at_speed < WHEEL_SPEED_SAMPLES) {
		top_at_speed = 0;
	}
    // bottom wheel
	if (bottom_PID->OnTarget()) {
		bottom_at_speed++;
	} else if (bottom_at_speed < WHEEL_SPEED_SAMPLES) {
		bottom_at_speed = 0;
	}
 
    // when both have been at speed long enough, we can shoot
    if (top_at_speed >= WHEEL_SPEED_SAMPLES && bottom_at_speed >= WHEEL_SPEED_SAMPLES)
    {
#if SHOOTER_LOGGING
		double cur_time = GetTime();
    	printf("AtSpeed is true %f\n", cur_time - at_speed_time);
#endif
    	return true;
    }
    
    return false;
}

float Shooter::getTopSpeed() {
	// returns the portion of total speed, such as the PID loop recieves
    return top_encoder->getLastSpeed();
} 
float Shooter::getBottomSpeed() {
    return bottom_encoder->getLastSpeed();
}

int Shooter::getTopEnc() {
    // returns total counts of the top encoder
    return top_encoder->getTotalCount();
} 
int Shooter::getBottomEnc() {
    return bottom_encoder->getTotalCount();
}

double Shooter::getTopRev() {
    return top_encoder->getRevPerSec();
} 
double Shooter::getBottomRev() {
    return bottom_encoder->getRevPerSec();
}

bool Shooter::isReady() {
	if ( atSpeed() && turretFinished() && store->isReady()) {
		return true;
	}
	return false;
}

void Shooter::shooterEnable() {
	shooter_enabled = true;

    if(!wheels_broken)
    {
    	// since wheels are just coming back on, then re-wait
    	top_at_speed = 0;
    	bottom_at_speed = 0;
    	
		top_PID->SetSetpoint(top_speed);
        top_PID->Enable();
		    
	    bottom_PID->SetSetpoint(bottom_speed);
        bottom_PID->Enable();

        at_speed_time = GetTime();
    } else {
    	bottom_jaguar->Set(bottom_speed);
    	top_jaguar->Set(top_speed);
    }
}
void Shooter::shooterDisable() {
	shooter_enabled = false;
	// When you have continuous speed, do not PIDs slow down
	if (!speed_variable_mode) return;
	
    if(!wheels_broken)
    {
		top_PID->Disable();
        top_PID->Reset();// to clear history
		    
		bottom_PID->Disable();
        bottom_PID->Reset();

    } else {
    	bottom_jaguar->Set(0);
    	top_jaguar->Set(0);
    }
}

void Shooter::setLoggingEnabled(bool en) {
#if SHOOTER_LOGGING
    top_logger->setEnabled(en);
    bottom_logger->setEnabled(en);
#endif
}


//------------------------------- TURRET ---------------------------------------------

void Shooter::setTurretBroken(bool broken) {
	broken_turret = broken;
	if (broken && turret_PID->IsEnabled()) {
	    turret_PID->Disable();
	    turret_jaguar->Set(0);
	}
}

void Shooter::setTurretAngle(float x) {
    
    turret_target_angle = x;
    turret_move_manual = false;    
}

bool Shooter::turretFinished() {
	// YAda
    if (broken_turret)
        return true;
    else
        return turretOnTarget();
}

void Shooter::moveTurret(float rotvel) {
	// tell motor to rotate with speed x proportional to rotvel
	turret_rotvel = rotvel;
	turret_move_manual = true;
}

void Shooter::stopTurret() {
	moveTurret(0);
}

float Shooter::getTurretDesiredAngle() {
    return turret_target_angle;
}

float Shooter::getTurretAngle() {
    return turretCountToAngle(turret_encoder->GetRaw());
}

int Shooter::getTurretEnc() {
    return turret_encoder->GetRaw();
}

double Shooter::getTurretEncNorm() {
    return turretNormalizeTicks(turret_encoder->GetRaw());
}

void Shooter::requestZero(){
	want_zero = true;
}

void Shooter::processTurret(){
	if(want_zero){
	    if (broken_turret) {
	        want_zero = false;
	    } else {
	        
    		//Do zeroing things
    	    if (turret_PID->IsEnabled())
    	        turret_PID->Disable();
    		findZero();
	    }
	} else {
		//Do normal things
	    
	    if (limitCW->GetVoltage() < 2) {
	        turret_encoder->Reset();
	        turret_encoder->Start();
	    }
	    
		//Check limits
		if(turret_move_manual)
		{
			
			if (turret_PID->IsEnabled()) turret_PID->Disable();

			turret_jaguar->Set(checkLimits(turret_rotvel));
		} else {
			// angle defined as positive being leftward, zero being robot main axis forward
		    double newAngle = turretAngleToNormalizedTicks(turret_target_angle);
            turret_PID->SetSetpoint(newAngle);
            if (!turretOnTarget()) {
                if (!turret_PID->IsEnabled())
                    turret_PID->Enable(); // make the PID active
            } else if (turret_PID->IsEnabled()) {
                // Stop it using Reset(), which also Disable()'s
                turret_PID->Reset();
            }
#if TURRET_DEBUG
		    static int limit = 0;
		    if ((limit++ % 20) == 0)
			    printf("Current PID output: %f want angle: %f setpoint: %f last measure: %f error: %f\n", (double)turret_PID->Get(), turret_target_angle, (double)turret_PID->GetSetpoint(), (double)turretNormalizeTicks(turret_encoder->getCount()), (double)turret_PID->GetError());
#endif          
		}
	}
}

float Shooter::checkLimits(float requestSpeed){
	if((limitCW->GetVoltage() < 2) && requestSpeed > 0)
		return 0;
	else if((limitCCW->GetVoltage() < 2) && requestSpeed < 0)
		return 0;
	else
		return requestSpeed;
}

float Shooter::getLimitCW(){
	return limitCW->GetVoltage();
}

float Shooter::getLimitCCW(){
	return limitCCW->GetVoltage();
}

void Shooter::findZero(){
	turret_jaguar->Set(0.4);
	if(limitCW->GetVoltage() < 2){
		turret_encoder->Reset();
		turret_encoder->Start();
		want_zero = false;
	}
}

double Shooter::getTurretError() {
    return turret_PID->GetSetpoint() - turret_encoder->PIDGet();
}

// Use this instead of turret_PID->OnTarget() or GetError() to avoid races
// between PID thread and main thread
bool Shooter::turretOnTarget() {
    double d = fabs(getTurretError());
    return d < TURRET_TOLERANCE;
}
    

static double turretNormalizeTicks(int ticks) {
    int middlecount = TURRET_MAX_COUNTS/2;
    // Map middlecount count-> 0 normalized
    //     0 count -> -1
    //     TURRET_MAX_COUNTS -> 1
    
    ticks -= middlecount;      // Center about 0
    double d = (double)ticks / middlecount;  // normalize
    return d;
}

static double turretCountToAngle(int count) {
    //actual angle is raw encoder counts with PID zero offset dialed in
    double d = turretNormalizeTicks(count);         // Normalize about -1 -> 0 -> 1
    d *= (TURRET_MAX_DEG - TURRET_MIN_DEG) / 2.0;     // -half -> half degree sweep
    d += TURRET_OFFSET_DEG;                         // -half+offset -> half+offset
    return d;
}

static double turretAngleToNormalizedTicks(double angle) {
    angle -= TURRET_OFFSET_DEG;
    angle /= (TURRET_MAX_DEG - TURRET_MIN_DEG) / 2.0;
    return angle;
}

