#ifndef SMART_SHOOTER
#define SMART_SHOOTER

#include "WPILib.h"
#include "Storage.h"
#include "Shooter.h"
#include "RPS.h"
#include "Camera.h"

//some #defines for targeting calculations....
// lots more to be found in Camera.h and RPS.h
/* TODO tweak this value */
#define CAMERA_TRIES_LIMIT 5
// weights for camera vs. RPS for angle and distance TODO tun(ing fork) it!!
                                                // DOOOONNNNNNGGGGGGGGGGGG......
#define WT_A_CAM 1/2.0
#define WT_A_RPS 1-WT_A_CAM
#define WT_D_CAM 1/2.0
#define WT_D_RPS 1-WT_D_CAM

// lookup table constants: TODO tune!
// minimum distance from shooter to basket
// slt[h,0] means not 0 in. from basket, but x0 inches
#define X0 36
// step size in lookup table: slt[h,0] vs slt[h,1]
// is an increase in distance of DX inches; i.e. 
// d index 1 is not x0+1 but x0+dx
#define DX 12
// longest shot we care to take (in inches)
#define MAX_SHOT 30*12
// size of lookup table; number of distance indices
// (array is slt[3][SLT_DSIZE])
#define SLT_DSIZE (MAX_SHOT-X0)/DX
// ^^ [currently 27]
// minimum and maximum cam servo angle
// these are values passed to and from cam->Get/SetTilt
// angle "0" is actually +50, "90" actually -39
// with respect to horizontal
#define MIN_ANGLE 0
#define MAX_ANGLE 90
// resulting size of rough speed lookup table
#define RSLT_SIZE (MAX_ANGLE-MIN_ANGLE)

class SmartShooter : public Shooter {
public:
	// see SetTargetMode for explanation of TM enum
	typedef enum TargetMode { off, camera_only, camera_and_rps, rps_only };
	
	SmartShooter(Storage *s);
	virtual ~SmartShooter();

	// Overrides the original; what it does now:
	//   if not targeting: just proceeds to call original (Shooter.cpp) version!
	//   but if targeting:
	//       asks Targeter for desired shooter speed and angle, then
	//       uses Shooter methods to set these values as necessary
	//       prepares for shot (actual shot will occur in superclass method)
	virtual void processShooter();
	
	// allow drivers to disable targeting
	//    t=true: turret auto-targets constantly; camera angle is locked (to
	//            prevent picture shake)
	//    t=false: it doesn't
	void setTargeting(bool t);
	// set TargetMode (i.e. how we find out angle and speed to use for the shot)
	//  -- using camera only; RPS only; or both together
	void SetTargetMode(TargetMode m);
    /* TODO how will this interface with the driver controls? */
	// tells SmartShooter which hoop we're aiming for 
	// (none, top, center-left, center-right, bottom)
	void SetTargetHoop(Camera::TargetIdentity t);
	
	// are we aimed, up to speed, and ready to shoot? [may be useful]
	// also returns false if not targeting, since then we don't have a target!
	bool onTarget();
	// tells Controls if targeting has failed, so alternative actions can
	// be taken (returns true if there was failure, false if we're still good) 
	bool targetingFailed();
	
	// tilt the camera (angle is degrees from horizontal, up is positive)
	//     if targeting, lock camera angle and don't allow further tilts
	void tiltCamera(float angle);
	// this function currently for debug purposes ONLY; get camera angle
	//    (same conventions as above)
	float getCamTilt();
	// for autonomous: have we reached the desired camera angle...
	// true = yes, false = no
	bool camTiltDone();
	
	// take camera angle and use it to get a rough estimate of motor
	// speed required for the shot (sent to dashboard for driver use)
	// Intent: drivers center top target in the camera vertically; then
	//         the dashboard displays the rough target speed; then
	//         drivers set speed to this value and fire.
	// Note: as of now, use ONLY for top hoop! performance not
	//      guaranteed for mid or low hoops.
	float getRoughTargetSpeed();
	
private:
	// calculation function; calculate values with respect to shooter mechanism
	// (i.e. distance is from center of shooter to hoop center; angle required
	//  CHANGE in turret angle in degrees)
	void CalcTargetDistanceAndAngle(float* distance, float* angle);
	// quick function to get height of target given
	
	RobotPositioningSystem *rps;
	Camera *cam;
	
	// what are we using to aim?
	TargetMode mode;
	// which one of the hoops are we aiming for?
	Camera::TargetIdentity target_hoop;
	//do we want the turret to auto-target? 
	bool targeting;
	// have we convincingly identified our position?
	bool positioned;
	// have we failed at finding targets (or was something too
	// broken for us to succeed?)
	bool failure;
	
	// targeting/camera overhead
	int camera_tries; // how many times have we unsuccessfully tried to find
					 	// the target in the camera image?
	
	// rough speed lookup table for top hoop (based on camera angle)
	//range is from -x degrees to +y degrees; index in lookup
	// table is 0 to 100 (i.e., -x+x to y+x)
	static float rough_speed_lookup_table[RSLT_SIZE];
	
	// targeting speed lookup table (based on calculated distance
	// and given target height); slt[0] is for bottom basket,
	// slt[1] is for middle baskets, slt[2] is for top basket
	// distance index 0 is X0 inches from basket; each higher
	// index represents a step of DX inches further from basket,
	// up to MAX_SHOT inches at distance index SLT_DSIZE
	static float speed_lookup_table[3][SLT_DSIZE];
};
#endif /*SMART_SHOOTER*/
