////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	The order for processing an image:
//		1. ConvertRgbToMono
//			- looks at given RGBValue image
//			- fills in monoImage
//			- fills in histogram with num pixels in each row of the monoImage
//			- for PC_TEST_APP, with erase background from given RGBValue image
//
//		2. CalcTargetRows
//			- looks at histogram
//			- fills in rowInfo[], but skips rows without enough pixels
//
//		3. CalcTargetColumns
//			- calls CreateRowHisto() and then FindTargets() for each row
//				CreateRowHisto():
//					- looks at monoImage and rowInfo[]
//					- fills in histogram with num pixel in each column within given row
//				FindTargets():
//					- looks at historgram and rowInfo[]
//					- fills in targetInfo[], will 'ignore' gaps ('merge' targets close together) and skip ones without enough pixels
//					- for PC_TEST_APP, fills in targetInfoTest[] and targetInfoTestG[]
//
//		4. CalcTargets
//			- calls FineTune() and CalcHeightAndTopEdge() for each target
//				FineTure():
//					- looks at targetInfo[] and monoImage
//					- uses historgram internally
//					- fills in targetInfo[].maxWidth/avgWidth
//					- fills in 'skew' of left edge
//					- updates targettInfo[].left/right/top/bottom to fit target exactly
//				CalcHeightAndTopEdge();
//					- looks at targetInfo[] and monoImage
//					- uses historgram internally
//					- fills in targetInfo[].maxHeight/avgHeight
//					- fills in 'skew' of top edge
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include "WPILib.h"
#include "ThunderVision.h"

ThunderVision::ThunderVision()
{
	Init();
}

ThunderVision::~ThunderVision()
{
}

// initialize member variables
void ThunderVision::Init()
{
	memset(histogram, 0, sizeof(histogram));
	memset(monoImage, 0, sizeof(monoImage));
	targets = 0;
	rows = 0;

	// for PC test app only
	#ifdef PC_TEST_APP
	targetsTest = 0;
	targetsTestG = 0;
	#endif // PC_TEST_APP
}

// will process the entire image all at once (i.e. run STEPs 1 through 4)
int ThunderVision::ProcessImage(RGBValue *image)
{
	ConvertRgbToMono(image);					// STEP 1
	CalcTargetRows();							// STEP 2
	CalcTargetColumns();						// STEP 3
	CalcTargets();								// STEP 4
	return (targets);
}

// convert the given RGB camera image to a monochrome image (i.e. pull out the green)
void ThunderVision::ConvertRgbToMono(RGBValue *image)
{
	unsigned char *mono;
	int x;
	int y;

	Init();

	// process each row
	mono = monoImage;
	for (y = 0; y < 480; y++)
	{
		// process each pixel in the row
		for (x = 0; x < 640; x++, mono++, image++)
		{
			if (   (image->G >= GREEN_THRESHOLD)			// green bright enough
				&& (image->R < (image->G - RED_DELTA))		// red far enough from green (need bluish-green)
				&& (image->B >= BLUE_THRESHOLD)				// blue bright enough (stay away from browns/blacks)
				&& (image->B <= (image->G + BLUE_DELTA))	// blue not too much brigther than green (stay away from blues)
				)
			{
				*mono = 255;
				histogram[y]++;								// keep track of row that has a pixel
			}
#ifdef PC_TEST_APP
			else
			{
				// for test app, wipe out the background
				image->R = image->G = image->B = 0;
			}
#endif // PC_TEST_APP
		}
	}
}

// look at each row to see if it has any pixels in it.
// any successive lines (with pixels) will be considered part of the same target
void ThunderVision::CalcTargetRows()
{
	int y;
	int i;
	int *histo;

	// look at each row
	rows = 0;					// no rows yet
	y = 0;						// start at first row
	histo = histogram;			// start at first row (using a pointer is faster at runtime)
	while ((y < 480) && (rows < MAX_ROWS))
	{
		// find first pixel
		while ((y < 480) && (*histo == 0))
		{
			y++;
			histo++;
		}
		if (y == 480)					// make sure we found a pixel. if not, go to next row
		{
			break;
		}
		rowInfo[rows].top = y;			// remember top-most row that has pixels

		// find the last pixel and keep count of how many total pixels there are
		i = 0;
		while (y < 480 && (*histo != 0))
		{
			i += *histo;
			y++;
			histo++;
		}
		rowInfo[rows].bottom = (y - 1);	// remember bottom-most row that has pixels

		// to help get rid of noise and diagonal line, we will only keep
		// this set if the average number of pixels is high enough
		if ((i / (rowInfo[rows].bottom - rowInfo[rows].top + 1)) > MIN_PIXELS_ROW)
		{
			rows++;
		}
	}
}

// look at each row (from CalcTargetRows()) and find all each box
void ThunderVision::CalcTargetColumns()
{
	int row;

	targets = 0;					// no targets yet
	#ifdef PC_TEST_APP
	targetsTest = 0;
	targetsTestG = 0;
	#endif // PC_TEST_APP

	// process each target row
	for (row = 0; row < rows; row++)
	{
		// create column histogram for the target row
		CreateRowHisto(row);

		// add targets to end of list
		// use 1/2 the height of the row as the amount of allowed 'gap' (i.e. ball/hoop)
		FindTargets(targets, row);
	}
}

// look at each target from CalcTargetColumns() and 'fine tune' it and get final values
void ThunderVision::CalcTargets()
{
	int target;

	for (target = 0; target < targets; target++)
	{
		FineTune(&targetInfo[target]);
		CalcHeightAndTopEdge(&targetInfo[target]);
	}
}

void ThunderVision::FineTune(TARGET_INFO *targetInfo)
{
	int x;
	int y;
	int left;
	int leftTmp;
	int right;
	int rightTmp;
	int top;
	int bottom;
	bool found = false;
	unsigned char *mono;

	memset(histogram, 0, sizeof(histogram));
	targetInfo->avgWidth = 0;
	targetInfo->maxWidth = 0;
	left = 641;
	right = 0;
	top = -1;
	for (y = targetInfo->top; y <= targetInfo->bottom; y++)
	{
		leftTmp = 641;					// track left-most pixel in this row
		rightTmp = 0;					// track right-most pixel in this row
		found = false;					// no pixels in this row yet
		mono = &monoImage[(y * 640) + targetInfo->left];
		for (x = targetInfo->left; x <= targetInfo->right; x++, mono++)
		{
			// if pixel is set, then update tracking variables
			if (*mono)
			{
				if (!found)
				{
					// it is the first found pixel, so save as left-most in this row
					leftTmp = x;
					histogram[y] = x;
				}
				rightTmp = x;			// any pixel with end up being the right-most pixel
				found = true;			// at least one pixel
			}
		}

		// if we found at least one pixel, then update overall tracking values
		if (found)
		{
			if (leftTmp < left)			// if this row's left is more left that prior row's, save it
			{
				left = leftTmp;
			}
			if (rightTmp > right)		// if this row's right is more right that prior row's, save it
			{
				right = rightTmp;
			}
			if (top == -1)				// if top not set yet, then save it
			{
				top = y;
			}
			bottom = y;					// any pixel in row ends up being the bottom-most row

			// keep track of max width and total width (so we can do average)
			if ((rightTmp - leftTmp + 1) > targetInfo->maxWidth)
			{
				targetInfo->maxWidth = (rightTmp - leftTmp + 1);
			}
			targetInfo->avgWidth += (rightTmp - leftTmp + 1);
		}
	}

	// save updated values
	targetInfo->left = left;
	targetInfo->right = right;
	targetInfo->top = top;
	targetInfo->bottom = bottom;
	targetInfo->avgWidth /= (bottom - top + 1);

	// determine the 'skew' of the left edge
	// we only want to look at the middle of the image (to avoid the angles of the top & bottom)
	x = 0;
	y = (targetInfo->bottom - targetInfo->top + 1) >> 2;	// get 1/4 of the total height
	rightTmp = targetInfo->bottom - y;						// end 1/4 from the bottom
	y += targetInfo->top;									// start 1/4 from the top
	leftTmp = histogram[y];									// initialize with first value
	for (; y <= rightTmp; y++)
	{
		// if value is greater than the previous, then the line
		// going down; otherise the line is going up
		if (histogram[y] > leftTmp)
		{
			x--;							// moved down
		}
		else if (histogram[y] < leftTmp)
		{
			x++;							// moved up
		}
		leftTmp = histogram[y];				// make current val the previous for next loop
	}
	targetInfo->edgeLeft = x;
}

void ThunderVision::CalcHeightAndTopEdge(TARGET_INFO *targetInfo)
{
	int x;
	int y;
	int topTmp;
	int bottomTmp;
	bool found = false;
	unsigned char *mono;

	targetInfo->avgHeight = 0;
	targetInfo->maxHeight = 0;
	memset(histogram, 0, sizeof(histogram));
	for (x = targetInfo->left; x <= targetInfo->right; x++)
	{
		found = false;					// no pixels in this row yet
		mono = &monoImage[(targetInfo->top * 640) + x];
		for (y = targetInfo->top; y <= targetInfo->bottom; y++, mono += 640)
		{
			// if pixel is set, then update tracking variables
			if (*mono)
			{
				if (!found)
				{
					// it is the first found pixel, so save as top-most in this row
					topTmp = y;
					histogram[x] = y;
				}
				bottomTmp = y;			// any pixel with end up being the bottom-most pixel
				found = true;			// at least one pixel
			}
		}

		// if we found at least one pixel, then update overall tracking values
		if (found)
		{
			// keep track of max height and the total height (so we can do average)
			if ((bottomTmp - topTmp + 1) > targetInfo->maxHeight)
			{
				targetInfo->maxHeight = (bottomTmp - topTmp + 1);
			}
			targetInfo->avgHeight += (bottomTmp - topTmp + 1);
		}
	}

	// calc average
	targetInfo->avgHeight /= (targetInfo->right - targetInfo->left + 1);

	// determine the 'skew' of the top edge
	// we only want to look at the middle of the image (to avoid the angles of the top & bottom)
	y = 0;
	x = (targetInfo->right - targetInfo->left + 1) >> 2;	// get 1/4 of the total width
	bottomTmp = targetInfo->right - x;						// end 1/4 from the right side
	x += targetInfo->left;									// start 1/4 from the left side
	topTmp = histogram[x];									// initialize with first value
	for (; x <= bottomTmp; x++)
	{
		// if value is greater than the previous, then the line going to right;
		// otherise the line is going to the left
		if (histogram[x] > topTmp)
		{
			y++;							// moved to the right
		}
		else if (histogram[x] < topTmp)
		{
			y--;							// moved to the left
		}
		topTmp = histogram[x];				// make current val the previous for next loop
	}
	targetInfo->edgeTop = y;
}

// gets the targets
//
// targetInfo		is a pointer to a list of targets (pass in NULL to get the total)
// numEntires		indicates how many TARGET_INFO structures that targetInfo points to
// returns			the number of targets copied
int ThunderVision::GetTargetInfo(TARGET_INFO *targetInfoList, int numEntries)
{
	int i;

	// if call just wants the total, then return it
	if (targetInfoList == NULL)
	{
		return (targets);
	}

	// copy what they want, but not more than want we have
	i = numEntries;
	if (i > targets)
	{
		i = targets;
	}
	memcpy(targetInfoList, targetInfo, sizeof(targetInfo[0]) * i);
	return (i);
}

// create column histogram for the given target row
void ThunderVision::CreateRowHisto(int row)
{
	unsigned char *mono;
	int x;
	int y;

	// reset histo
	memset(histogram, 0, sizeof(histogram));

	// process each row within a single target row
	mono = monoImage;
	for (y = rowInfo[row].top; y <= rowInfo[row].bottom; y++)
	{
		// process each pixel in the row
		mono = &monoImage[y * 640];
		for (x = 0; x < 640; x++, mono++)
		{
			if (*mono)
			{
				histogram[x]++;	// keep track of column that has a pixel
			}
		}
	}
}

// look at row's histogram and find all the targets (just getting 'width' of target)
void ThunderVision::FindTargets(int targetFirstInRow, int row)
{
	int x;
	int i;
	int *histo;
	int gap;

	// determine how much gap between targets is allowed (i.e. ball/hoop)
	// if a target is this close to the previous target, then it will be
	// appened to the previous target
	gap = (rowInfo[row].bottom - rowInfo[row].top + 1) >> 1;	// half of the hieght

	// look at each row
	x = 0;						// start at first col
	histo = histogram;			// start at first col (using a pointer is faster at runtime)
	while ((x < 640) && (targets < MAX_TARGETS))
	{
		// find first pixel
		while ((x < 640) && (*histo == 0))
		{
			x++;
			histo++;
		}
		if (x == 640)					// make sure we found a pixel. if not, go to next col
		{
			break;
		}
		targetInfo[targets].row = row;						// remember row target is in
		targetInfo[targets].left = x;						// remember left-most col that has pixels
		targetInfo[targets].top = rowInfo[row].top;			// remember vertical placement
		targetInfo[targets].bottom = rowInfo[row].bottom;	// remember vertical placement
		targetInfo[targets].avgWidth = 0;					// 0 it since we add to this when appending

		// find the last pixel and keep count of how many total pixels there are
		i = 0;
		while (x < 640 && (*histo != 0))
		{
			i += *histo;
			x++;
			histo++;
		}

		#ifdef PC_TEST_APP
		if (targetsTest < MAX_TARGETS_TEST)
		{
			memcpy(&targetInfoTest[targetsTest], &targetInfo[targets], sizeof(targetInfoTest[0]));
			targetInfoTest[targetsTest].right = (x - 1);
			targetsTest++;
		}
		#endif // PC_TEST_APP

		// if we already have a target, then see if this one is close to
		// the previous one. this handles gaps caused by balls and hoop
		if (   (targets > targetFirstInRow)
			&& (targetInfo[targets].left < (targetInfo[targets - 1].right + gap))
			)
		{
			targets--;								// append to previous target
		}

		targetInfo[targets].right = (x - 1);		// remember right-most col that has pixels
		targetInfo[targets].avgWidth += i;
		targets++;
	}

	// now that we found all the targets for this row, we need
	// to strip out any that are too small. do this pre row so we
	// don't overfill targetInfo
	for (x = targetFirstInRow; x < targets; x++)
	{
		#ifdef PC_TEST_APP
		if (targetsTestG < MAX_TARGETS_TEST_G)
		{
			memcpy(&targetInfoTestG[targetsTestG], &targetInfo[x], sizeof(targetInfoTestG[0]));
			targetsTestG++;
		}
		#endif // PC_TEST_APP

		// to help get rid of noise and diagonal line, we will only keep
		// this set if the average number of pixels is high enough
		if ((targetInfo[x].avgWidth / (targetInfo[x].bottom - targetInfo[x].top + 1)) <= MIN_PIXELS_ROW)
		{
			for (i = x + 1; i < targets; i++)
			{
				memcpy(&targetInfo[i - 1], &targetInfo[i], sizeof(targetInfo[0]));
			}
			x--;
			targets--;
		}
	}
}

// the following is only used by the PC test app
#ifdef PC_TEST_APP
void ThunderVision::GetTargetRowImage(RGBValue *image)
{
	int row;
	int y;
	int x;

	for (row = 0; row < rows; row++)
	{
		for (y = rowInfo[row].top; y <= rowInfo[row].bottom; y++)
		{
			for (x = 0; x < 640; x++)
			{
				// leave in original boxes
				if (image[(y * 640) + x].G == 0)
				{
					// 1st row is RED, 2nd row if BLUE, 3rd row is GREEN, 4th row is WHITE, etc.
					image[(y * 640) + x].R = (((row & 0x03) == 0) || ((row & 0x03) == 3)) ? 255 : 0;
					image[(y * 640) + x].G = (((row & 0x03) == 1) || ((row & 0x03) == 3)) ? 255 : 0;
					image[(y * 640) + x].B = (((row & 0x03) == 2) || ((row & 0x03) == 3)) ? 255 : 0;
				}
			}
		}
	}
}

void ThunderVision::GetTargetImageWithGap(RGBValue *image)
{
	GetTargetImage(image, targetsTest, targetInfoTest);
}
void ThunderVision::GetTargetImageNoGap(RGBValue *image)
{
	GetTargetImage(image, targetsTestG, targetInfoTestG);
}
void ThunderVision::GetTargetImageNoNoise(RGBValue *image)
{
	GetTargetImage(image, targets, targetInfo);
}
void ThunderVision::GetTargetImageTight(RGBValue *image)
{
	GetTargetImage(image, targets, targetInfo);
}
void ThunderVision::GetTargetImage(RGBValue *image, int maxTargets, TARGET_INFO *targetInfo)
{
	int target;
	int y;
	int x;

	for (target = 0; target < maxTargets; target++)
	{
		for (y = targetInfo[target].top; y <= targetInfo[target].bottom; y++)
		{
			for (x = targetInfo[target].left; x <= targetInfo[target].right; x++)
			{
				// leave in original boxes
				if (image[(y * 640) + x].G == 0)
				{
					// 1st target is RED, 2nd target if BLUE, 3rd target is GREEN, 4th target is WHITE, etc.
					image[(y * 640) + x].R = (((target & 0x03) == 0) || ((target & 0x03) == 3)) ? 255 : 0;
					image[(y * 640) + x].G = (((target & 0x03) == 1) || ((target & 0x03) == 3)) ? 255 : 0;
					image[(y * 640) + x].B = (((target & 0x03) == 2) || ((target & 0x03) == 3)) ? 255 : 0;
				}
			}
		}
	}
}

#endif // PC_TEST_APP
